{ writeTextFile
, stdenv
, lib
, createCredentials

# Path to the rc.subr script
, rcSubr ? "/etc/rc.subr"
# Specifies which command are builtin. This is to determine which extra commands may have been provided.
, builtinCommands ? [ "start" "stop" "reload" "restart" "status" "poll" "rcvar" ]
# Specifies the default signal used for reloading a process
, defaultReloadSignal ? "HUP"
# Specifies the default signal used for stopping a process
, defaultStopSignal ? "TERM"
# Default run time directory where PID files are stored
, runtimeDir ? "/var/run"
# Directory in which log files are stored
, logDir ? "/var/log"
# Directory in which temp files are stored
, tmpDir ? "/tmp"
# Specifies whether user changing functionality should be disabled or not
, forceDisableUserChange ? false
}:

{
# A name that identifies the process instance
name ? instanceName
# The variable suffix that indicates whether a service has been enabled or not.
, rcvar ? "enabled"
# An attribute set defining default values for configuration environment variables
, rcvarsDefaults ? {}
# The command that executes a daemon
, command ? null
# The command-line parameters passed to the command
, commandArgs ? []
# Specifies whether the command daemonizes or not. If the command is not a daemon, it gets daemonized by the generator
, commandIsDaemon ? true
# Whether to disable logging the daemon process' output
, enableDaemonOutputLogging ? true
# Specifies the signal that needs to be sent to reload a process
, reloadSignal ? "HUP"
# Specifies the signal that needs to be sent to stop a process
, stopSignal ? "TERM"
# Specifies which packages need to be in the PATH
, environment ? {}
# An attribute set specifying arbitrary environment variables
, path ? []
# A name that uniquely identifies each process instance. It is used to generate a unique PID file.
, instanceName ? null
# Path to a PID file that the system should use to manage the process. If null, it will use a default path.
, pidFile ? null
# If not null, the nice level be changed before executing any activities
, nice ? null
# If not null, the umask will be changed before executing any activities
, umask ? null
# If not null, the current working directory will be changed before executing any activities
, directory ? null
# Specifies as which user the process should run. If null, the user privileges will not be changed.
, user ? null
# FIB Routing Table number to run command with
, fib ? null
# Resource limits to apply to command. This will be passed as arguments to the limits utility.
, limits ? null
# Login class to use. Defaults to "daemon".
, loginClass ? null
# Protect command from being killed when swap space is exhausted. If "YES" is used, no child processes are protected.  If "ALL", protect all child processes.
, oomProtect ? null
# Defines files that must be readable before running the start method
, requiredFiles ? []
# Defines directories that must exist before running the start method
, requiredDirs ? []
# Defines external environment variables this script depends on
, requiredVars ? []
# Perform checkyesno on each of the list variables before running the start method.
, requiredModules ? []
# Specifies the implementation of arbitrary commands
, commands ? {}
# If set to true the rc script accepts an arbitrary number of parameters. If set to false, it only accepts one.
, flexibleParameters ? false
# Specifies which feature the script requires. This is used by the rc init system to determine the proper activation order
, requires ? []
# Specifies which features the script provides. By default, it simply considers its name a feature.
, provides ? [ "${name}" ]
# Keywords to be display in the comments section
, keywords ? []
# A list of bsd rc scripts that this script depends on
, dependencies ? []
# Specifies which groups and users that need to be created.
, credentials ? {}
# Arbitrary commands executed after generating the configuration files
, postInstall ? ""
}:

assert command == null -> commands ? start && commands ? stop;
assert name != null;

let
  util = import ../util {
    inherit lib;
  };

  extraCommands = builtins.attrNames (removeAttrs commands builtinCommands);

  _user = util.determineUser {
    inherit user forceDisableUserChange;
  };

  outputLogFile = util.autoGenerateDaemonLogFilePath {
    inherit name instanceName logDir tmpDir enableDaemonOutputLogging;
    user = _user;
  };

  _command = if commandIsDaemon then command else "daemon";
  _commandArgs = if commandIsDaemon then commandArgs else
    lib.optionals (pidFile != null) [ "-p" pidFile ]
    ++ lib.optionals (outputLogFile != null) [ "-o" outputLogFile ]
    ++ [ command ]
    ++ commandArgs;

  _requires = map (dependency: dependency.name) dependencies ++ requires;

  envFile = if environment == {} then null else writeTextFile {
    name = "${name}-envfile";
    text = lib.concatMapStrings (name:
      let
        value = builtins.getAttr name environment;
      in
      ''${name}=${lib.escapeShellArg value}
      ''
    ) (builtins.attrNames environment);
  };

  pidFilesDir = util.determinePIDFilesDir {
    inherit user runtimeDir tmpDir; # We can't use _user because we want to keep the path convention the same
  };

  _pidFile = util.autoGeneratePIDFilePath {
    inherit pidFile instanceName pidFilesDir;
  };

  rcScript = writeTextFile {
    inherit name;
    executable = true;
    text = ''
      #!/bin/sh
    ''
    + lib.optionalString (provides != []) ''
      # PROVIDE: ${toString provides}
    ''
    + lib.optionalString (_requires != []) ''
      # REQUIRE: ${toString _requires}
    ''
    + lib.optionalString (keywords != []) ''
      # KEYWORD: ${toString keywords}
    '' +
    ''
      . ${rcSubr}

      name="${name}"
    '' + lib.optionalString (rcvar != null) ''
      rcvar=''${name}_${rcvar}
    ''
    + ''

      load_rc_config $name
      ${lib.concatMapStrings (rcvarName: ''
        : ''${name}_${rcvarName}:=${toString builtins.getAttr rcvarName rcvarsDefaults}
      '') (builtins.attrNames rcvarsDefaults)}

    ''
    + lib.optionalString (umask != null) ''
      umask ${umask}
    ''
    + lib.optionalString (_command != null) ''
      command=${_command}
    ''
    + lib.optionalString (_commandArgs != []) ''
      command_args="${lib.escapeShellArgs _commandArgs}"
    ''
    + lib.optionalString (requiredDirs != []) ''
      required_dirs="${toString requiredDirs}"
    ''
    + lib.optionalString (requiredFiles != []) ''
      required_files="${toString requiredFiles}"
    ''
    + lib.optionalString (requiredVars != []) ''
      required_vars="${toString requiredVars}"
    ''
    + lib.optionalString (requiredModules != []) ''
      required_modules="${toString requiredModules}"
    ''
    + lib.optionalString (_pidFile != null) ''
      pidfile="${_pidFile}"
    ''
    + lib.optionalString (reloadSignal != defaultReloadSignal) ''
      sig_reload="${reloadSignal}"
    ''
    + lib.optionalString (stopSignal != defaultStopSignal) ''
      sig_stop="${stopSignal}"
    ''
    + lib.optionalString (nice != null) ''
      ${name}_nice=${toString nice}
    ''
    + lib.optionalString (directory != null) ''
      ${name}_chdir=${directory}
    ''
    + lib.optionalString (_user != null) ''
      ${name}_user=${_user}
    ''
    + lib.optionalString (envFile != null) ''
      ${name}_env_file=${envFile}
    ''
    + lib.optionalString (fib != null) ''
      ${name}_fib=${fib}
    ''
    + lib.optionalString (limits != null) ''
      ${name}_limits=${limits}
    ''
    + lib.optionalString (loginClass != null) ''
      ${name}_login_class=${loginClass}
    ''
    + lib.optionalString (oomProtect != null) ''
      ${name}_oomprotect=${oomProtect}
    ''
    + lib.optionalString (extraCommands != []) ''
      extra_commands="${toString extraCommands}"
    ''
    + lib.concatMapStrings (commandName:
      let
        command = builtins.getAttr commandName commands;
      in
      lib.optionalString (command ? pre) ''${commandName}_precmd=''${name}_pre${commandName}
      ''
      + lib.optionalString (command ? implementation) ''${commandName}_cmd=''${name}_${commandName}
      ''
      + lib.optionalString (command ? post) ''${commandName}_postcmd=''${name}_post${commandName}
      ''
    ) (builtins.attrNames commands)
    + lib.optionalString (path != []) ''

      PATH=${util.composePathEnvVariable { inherit path; }}
      export PATH
    ''
    + "\n"
    + lib.concatMapStrings (commandName:
      let
        command = builtins.getAttr commandName commands;
      in
      ''
        ${lib.optionalString (command ? pre) ''
          ${name}_pre${commandName}()
          {
              ${command.pre}
          }
        ''}
        ${lib.optionalString (command ? implementation) ''
          ${name}_${commandName}()
          {
              ${command.implementation}
          }
        ''}
        ${lib.optionalString (command ? post) ''
          ${name}_post${commandName}()
          {
              ${command.post}
            }
        ''}
      ''
    ) (builtins.attrNames commands)
    + ''
      run_rc_command "${if flexibleParameters then "$@" else "$1"}"
    '';
  };

  credentialsSpec = createCredentials credentials;
in
stdenv.mkDerivation {
  inherit name;

  buildCommand = ''
    mkdir -p $out/etc/rc.d
    cd $out/etc/rc.d
    ln -s ${rcScript} ${name}
    ln -s ${credentialsSpec}/dysnomia-support $out/dysnomia-support

    cd $TMPDIR
    ${postInstall}
  '';
}
